/*
 * -------------------------------------------------------------
 *
 * This file contains function which are not only useful with
 * the MV-1000
 *
 * Copyright (C)  1996 Jochen Karrer 
 *
 * Released under the terms of the GPL.
 *
 * ------------------ ------------------------------------------
 *
 */

#include <sys/types.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <jpeglib.h>
#include <setjmp.h>
#include <X11/Xlib.h>
#include <X11/X.h>
#include "fglib.h"

/*
 * ---------------------------------------------------------------------
 *
 *  write a jpeg file using libjpeg
 *  The image_width is aligned to 4 Pixels when you are using
 *  the MV1000!  Call it with the size you got from a query to the
 *  library:
 *  MV1_InquireFrame(&frame,board);
 *  image_width  = frame.dx;
 *  image_height = frame.dy;
 *
 * --------------------------------------------------------------------
 */

int
FG_WriteJPEGFile (      JSAMPLE *image_buffer,
			char *comment,
                        int image_width,
                        int image_height,
                        FILE * outfile, 
                        int quality)
{
  	struct jpeg_compress_struct cinfo;
  	struct jpeg_error_mgr jerr;
	JSAMPROW row_pointer[1];      /* pointer to JSAMPLE row[s] */
	int row_stride;               /* physical row width in image buffer */
	int i;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);
	jpeg_stdio_dest(&cinfo, outfile);

	cinfo.image_width = image_width; 
	cinfo.image_height = image_height;
	cinfo.input_components = 1;                   /* # of color components per pixel */
	cinfo.in_color_space = JCS_GRAYSCALE;         /* colorspace of input image */
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, quality, TRUE /* limit to baseline-JPEG values */);
	jpeg_start_compress(&cinfo, TRUE);

	if(comment[0]) 
		jpeg_write_marker(&cinfo,JPEG_COM,comment,strlen(comment)-1);
	row_stride = image_width * 1; /* JSAMPLEs per row in image_buffer */

	for(i=0;i < cinfo.image_height;i++) {
  		row_pointer[0] = & image_buffer[ i * row_stride];
  		jpeg_write_scanlines(&cinfo, row_pointer, 1);
	} 

	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);
	return 0;

}
/*
 * ------------------------------------------------------------
 *
 * Read/Write PPM and PGM
 *
 * ------------------------------------------------------------
 */
int
FG_ReadPnm(FG_pnm *pnm,FILE *file) {
        int xsize,ysize,type,numcols;
        int c;
	fseek(file,0,SEEK_SET);
        fscanf(file,"P%d\n",&type);
        while((c=fgetc(file))=='#') {
                while( ((c=fgetc(file))!='\n') && (c!=EOF));
        }
        ungetc(c,file);
        fscanf(file,"%d %d\n",&xsize,&ysize);
        fscanf(file,"%d\n",&numcols);
        pnm->type = type;
        if(pnm->type == 6)
                pnm->pixelbytes = 3;
        else
                pnm->pixelbytes = 1;
        pnm->xsize = xsize;
        pnm->ysize = ysize;
        pnm->numcols = numcols;
//        fprintf(stderr,"xsize %d,ysize %d,bytes %d, cols %d\n",xsize,ysize,pnm->pixelbytes,numcols);

        if(xsize*ysize*pnm->pixelbytes < 5*1024*1024) {
                pnm->buf = malloc(pnm->xsize*pnm->ysize*pnm->pixelbytes);
                if(pnm->buf==NULL)
                        return -1;
        }
        fseek(file,-pnm->pixelbytes*pnm->xsize*pnm->ysize,SEEK_END);
        return fread(pnm->buf,pnm->pixelbytes,pnm->xsize*pnm->ysize,file);
}
int
FG_WritePnm(FG_pnm *pnm,FILE *file) {
	fseek(file,0,SEEK_SET);
	fprintf(file,"P%d\n",pnm->type);
	fprintf(file,"#hello\n");
	fprintf(file,"%d %d\n",pnm->xsize,pnm->ysize);
	fprintf(file,"%d\n",pnm->numcols);
	return fwrite(pnm->buf,pnm->pixelbytes,pnm->xsize*pnm->ysize,file);
}

/*
 * ------------------------------------------------------------------------------------
 *  
 * Help functions for filling a colormap that is similar to thw existing one.
 * useful for displaying on 8Bit-Pseudocolor Visuals
 * 
 * ------------------------------------------------------------------------------------
 */

static double 
calc_coldiff(XColor col1, XColor col2) {
        double diff;
        double red = abs(col1.red - col2.red);
        double green = abs(col1.green - col2.green);
        double blue = abs(col1.blue - col2.blue);
        //diff = sqrt(red*red+green*green+blue*blue)/3/65536;
        diff =red*red+green*green+blue*blue;
        return diff; 
}
int 
FG_TryGlobalCmap(Display *display,XColor tofind[],
		int num_cols,u_long pixeltab[]) {

	Screen *screen = DefaultScreenOfDisplay(display);
	Colormap cmap = DefaultColormapOfScreen (screen);
	int i;

	for (i=0; i< num_cols; i++) {
                tofind[i].flags = 0xff;
        	if (!XAllocColor (display,cmap,&tofind[i])) 
                	break;
        	pixeltab[i] = tofind[i].pixel;
     	}
	if(i < num_cols) {
		XFreeColors(display, cmap, pixeltab,i,0);
		return -1;
	}
	return 1;
}
/*
 * -----------------------------------------------------------------------------
 *
 * Create a colormap (newmap) which is very similar to the original color map
 * (In most cases the orignal map will be the DefaultColormapOfScreen (screen))
 * 
 * -----------------------------------------------------------------------------
 */
void
FG_FindSimilarCmap(Display *display,Colormap origmap,
			Colormap newmap,XColor tofind[],int num_cols,u_long pixeltab[]){
        XColor col[256];
        int usedtab[256];
        int i,j;
	static int first_call = 1;

        for(i=0;i<256;i++){
                col[i].pixel = i;
                usedtab[i] = -1;
        }

	// Only for creating entries for  colors in nearly empty colormaps
        if(first_call) {
                first_call=0;
                if(FG_TryGlobalCmap(display,tofind,
                        num_cols,pixeltab)>0) {
                        XFreeColors(display, DefaultColormapOfScreen(DefaultScreenOfDisplay(display)),
                                pixeltab,num_cols,0);
                }
        }

	
	
        XQueryColors(display,origmap,col,256);
        for(j=0;j<num_cols;j++){
                double best = 1e99;
                int bestfound = 0;
                for(i=0;i<256;i++) {
                        if((usedtab[i]<0) && (calc_coldiff(tofind[j],col[i]) < best) ) {
                                best = calc_coldiff(tofind[j],col[i]);
//                                printf("best %f\n",best);
                                bestfound = i;
                        }
                }
                usedtab[bestfound] = j; // mark used
                tofind[j].flags = 0xff;
                tofind[j].pixel = bestfound;
                pixeltab[j] = bestfound;
                XStoreColor(display,newmap,&tofind[j]);
        }
        for(i=255;i>=0;i--){
                if(usedtab[i]<0) {
                       XStoreColor(display,newmap,&(col[i]));
                }
        }
}

